from sqf.types import Keyword, Number, Array, Code, Type, Boolean, String, Namespace
from sqf.interpreter_types import WhileType, ForType, ForSpecType, SwitchType, IfType, ElseType, TryType, WithType
from sqf.expressions import BinaryExpression, UnaryExpression


class WhileExpression(UnaryExpression):
    """
    Catches `While {}` expression and stores it as a WhileType
    """
    def __init__(self):
        super().__init__(Keyword('while'), Code, WhileType, lambda v, i: v)


class WhileDoExpression(BinaryExpression):
    def __init__(self, action=None):
        super().__init__(WhileType, Keyword('do'), Code, None, action)


class ForSpecExpression(UnaryExpression):
    def __init__(self):
        super().__init__(Keyword('for'), Array, ForSpecType, lambda v, i: v)


class ForSpecDoExpression(BinaryExpression):
    def __init__(self, action=None):
        super().__init__(ForSpecType, Keyword('do'), Code, None, action)


class ForExpression(UnaryExpression):
    def __init__(self):
        super().__init__(Keyword('for'), String, ForType, lambda rhs, i: rhs)


class ForFromExpression(BinaryExpression):
    def __init__(self):
        super().__init__(ForType, Keyword('from'), Number, ForType,
                         lambda lhs, rhs, i: (lhs.variable, rhs))


class ForFromToExpression(BinaryExpression):
    def __init__(self):
        super().__init__(ForType, Keyword('to'), Number, ForType,
                         lambda lhs, rhs, i: (lhs.variable, lhs.from_, rhs))


class ForFromToStepExpression(BinaryExpression):
    def __init__(self):
        super().__init__(ForType, Keyword('step'), Number, ForType,
                         lambda lhs, rhs, i: (lhs.variable, lhs.from_, lhs.to, rhs))


class ForFromToDoExpression(BinaryExpression):
    def __init__(self, action=None):
        super().__init__(ForType, Keyword('do'), Code, None, action)


class ForEachExpression(BinaryExpression):
    def __init__(self, action=None):
        super().__init__(Code, Keyword('forEach'), Array, None, action)


class SwitchExpression(UnaryExpression):
    def __init__(self):
        super().__init__(Keyword('switch'), Type, SwitchType, lambda v, i: (self.keyword, v))


class CaseExpression(UnaryExpression):
    def __init__(self):
        super().__init__(Keyword('case'), Type, SwitchType, lambda v, i: (self.keyword, v))


class DefaultExpression(UnaryExpression):
    def __init__(self):
        super().__init__(Keyword('default'), Type, SwitchType, lambda v, i: (self.keyword, v))


class SwitchDoExpression(BinaryExpression):
    def __init__(self, action=None):
        super().__init__(SwitchType, Keyword('do'), Code, None, action)


class IfExpression(UnaryExpression):
    def __init__(self):
        super().__init__(Keyword('if'), Boolean, IfType, lambda v, i: v)


class IfThenExpression(BinaryExpression):
    def __init__(self, action=None):
        super().__init__(IfType, Keyword('then'), Code, None, action)


class ElseExpression(BinaryExpression):
    def __init__(self):
        super().__init__(Code, Keyword('else'), Code, ElseType, lambda lhs, rhs, i: (lhs, rhs))


class IfThenElseExpression(BinaryExpression):
    def __init__(self, action=None):
        super().__init__(IfType, Keyword('then'), ElseType, None, action)


class IfThenSpecExpression(BinaryExpression):
    def __init__(self, action=None):
        super().__init__(IfType, Keyword('then'), Array, None, action)


class IfThenExitWithExpression(BinaryExpression):
    def __init__(self, action=None):
        super().__init__(IfType, Keyword('exitwith'), Code, None, action)


class TryExpression(UnaryExpression):
    def __init__(self):
        super().__init__(Keyword('try'), Code, TryType, lambda v, i: v)


class TryCatchExpression(BinaryExpression):
    def __init__(self, action=None):
        super().__init__(TryType, Keyword('catch'), Code, None, action)


class WithExpression(UnaryExpression):
    def __init__(self):
        super().__init__(Keyword('with'), Namespace, WithType, lambda v, i: v)


class WithDoExpression(BinaryExpression):
    def __init__(self, action=None):
        if action is None:
            action = lambda lhs, rhs, i: i.execute_code(rhs, namespace_name=lhs.namespace.value)
        super().__init__(WithType, Keyword('do'), Code, None, action)


COMMON_EXPRESSIONS = [
    CaseExpression(),
    DefaultExpression(),

    WithExpression(),
    WithDoExpression(),

    TryExpression(),
    TryCatchExpression(),

    ForEachExpression(),
    WhileExpression(),
    WhileDoExpression(),

    ForExpression(),
    ForFromExpression(),
    ForFromToExpression(),
    ForFromToStepExpression(),
    ForFromToDoExpression(),
    ForSpecExpression(),
    ForSpecDoExpression(),
    SwitchExpression(),
    SwitchDoExpression(),

    IfExpression(),
    ElseExpression(),
    IfThenSpecExpression(),
    IfThenElseExpression(),
    IfThenExpression(),
    IfThenExitWithExpression(),

    UnaryExpression(Keyword('params'), Array, Boolean, lambda rhs_v, i: i.add_params(rhs_v)),
    BinaryExpression(Type, Keyword('params'), Array, Boolean, lambda lhs_v, rhs_v, i: i.add_params(rhs_v, lhs_v)),

    UnaryExpression(Keyword('call'), Code, None, lambda rhs_v, i: i.execute_code(rhs_v)),
    BinaryExpression(Type, Keyword('call'), Code, None, lambda lhs_v, rhs_v, i: i.execute_code(rhs_v, extra_scope={"_this": lhs_v})),
]



import math

from sqf.common_expressions import TryCatchExpression, ForEachExpression, \
    WhileDoExpression, ForFromToDoExpression, ForSpecDoExpression, SwitchDoExpression, \
    IfThenSpecExpression, IfThenElseExpression, IfThenExpression, IfThenExitWithExpression
from sqf.types import Keyword, Namespace, Number, Array, Code, Type, Boolean, String, Nothing, Variable
from sqf.exceptions import SQFParserError
from sqf.keywords import OP_ARITHMETIC, OP_COMPARISON, OP_LOGICAL
from sqf.expressions import BinaryExpression, UnaryExpression
from sqf.interpreter_types import SwitchType


OP_OPERATIONS = {
    # Arithmetic
    Keyword('+'): lambda x, y: x + y,
    Keyword('-'): lambda x, y: x - y,
    Keyword('*'): lambda x, y: x * y,
    Keyword('/'): lambda x, y: x / y,
    Keyword('%'): lambda x, y: x % y,
    Keyword('mod'): lambda x, y: x % y,
    Keyword('^'): lambda x, y: x ** y,
    Keyword('max'): lambda x, y: max(x, y),
    Keyword('floor'): lambda x: math.floor(x),

    # Comparison
    Keyword('=='): lambda x, y: x == y,
    Keyword('!='): lambda x, y: x != y,
    Keyword('<'): lambda x, y: x < y,
    Keyword('>'): lambda x, y: x < y,
    Keyword('<='): lambda x, y: x <= y,
    Keyword('>='): lambda x, y: x >= y,

    # Logical
    Keyword('&&'): lambda x, y: x and y,
    Keyword('and'): lambda x, y: x and y,
    Keyword('||'): lambda x, y: x or y,
    Keyword('or'): lambda x, y: x or y,
}


class ComparisonExpression(BinaryExpression):

    def __init__(self, op, lhs_rhs_type):
        assert(op in OP_COMPARISON)
        assert (issubclass(lhs_rhs_type, Type))
        super().__init__(lhs_rhs_type, op, lhs_rhs_type, Boolean, self._action)

    def _action(self, lhs, rhs, _):
        return OP_OPERATIONS[self.keyword](lhs.value, rhs.value)


class ArithmeticExpression(BinaryExpression):

    def __init__(self, op):
        assert (op in OP_ARITHMETIC)
        super().__init__(Number, op, Number, Number, self._action)

    def _action(self, lhs, rhs, _):
        return OP_OPERATIONS[self.keyword](lhs.value, rhs.value)


class LogicalExpression(BinaryExpression):
    def __init__(self, op, rhs_type):
        assert (op in OP_LOGICAL)
        assert (rhs_type in (Boolean, Code))
        super().__init__(Boolean, op, rhs_type, Boolean, self._action)

    def _action(self, lhs, rhs, interpreter):
        if isinstance(rhs, Code):
            result = interpreter.execute_code(rhs)
            if type(result) not in (Boolean, Nothing):
                interpreter.exception(SQFParserError(rhs.position, 'code return must be a Boolean (returns %s)' % type(result).__name__))
                return None
        else:
            result = rhs

        return OP_OPERATIONS[self.keyword](lhs.value, result.value)


class Action:
    def __init__(self, action):
        self.action = action

    def __call__(self, *args):
        result = None
        # interpreter = args[-1]
        all_args = args[:-1]
        return self.action(*all_args)


def _select(lhs, rhs, interpreter):
    index = int(round(rhs.value))
    try:
        return lhs[index]
    except IndexError:
        interpreter.exception(SQFParserError(
            lhs.position, 'selecting element %d of array of size %d' % (index, len(lhs))))


def _select_array(lhs, rhs, interpreter):
    start = rhs.value[0].value
    count = rhs.value[1].value

    if start > len(lhs.value):
        interpreter.exception(SQFParserError(lhs.position, 'Selecting element past size'))

    return lhs.value[start:start + count]


def _subtract_arrays(lhs, rhs):
    rhs_set = set([rhs_i.value for rhs_i in rhs.value])
    return [lhs_i for lhs_i in lhs if lhs_i.value not in rhs_set]


def _find(lhs_v, rhs_v):
    try:
        index = next(i for i, v in enumerate(lhs_v.value) if v == rhs_v)
    except StopIteration:
        index = -1
    return index


def _pushBack(lhs_v, rhs_v):
    lhs_v.append(rhs_v)
    return len(lhs_v.value) - 1


def _pushBackUnique(lhs_v, rhs_v):
    if rhs_v in lhs_v.value:
        return -1
    else:
        lhs_v.append(rhs_v)
        return len(lhs_v.value) - 1


def _setVariable(lhs_v, rhs_v, interpreter):
    namespace_name = lhs_v.value
    assert(isinstance(rhs_v, Array))

    if len(rhs_v) not in [2, 3]:
        interpreter.exception(SQFParserError(
            rhs_v.position, 'setVariable requires array of 2-3 elements (has %d)' % (len(rhs_v))))

    # get the variable name
    if not isinstance(rhs_v.value[0], (String, Nothing)):
        interpreter.exception(SQFParserError(
            rhs_v.value[0].position, 'setVariable array first element must be a string (is %s)' % type(rhs_v.value[0]).__name__))

    variable_name = rhs_v.value[0].value
    # get the value
    rhs_assignment = rhs_v.value[1]

    scope = interpreter.get_scope(variable_name, namespace_name)
    scope[variable_name] = rhs_assignment


def _getVariableString(lhs_v, rhs_v, interpreter):
    variable = Variable(rhs_v.value)
    variable.position = rhs_v.position
    return interpreter.value(variable, lhs_v.value)


def _getVariableArray(lhs_v, rhs_v, interpreter):
    # get the variable name
    if len(rhs_v) != 2:
        interpreter.exception(SQFParserError(
            rhs_v.position, 'getVariable requires array of 2 elements (has %d)' % (len(rhs_v))))

    if not isinstance(rhs_v.value[0], (String, Nothing)):
        interpreter.exception(SQFParserError(
            rhs_v.value[0].position, 'getVariable array first element must be a string (is %s)' % type(rhs_v.value[0]).__name__))

    variable = Variable(rhs_v.value[0].value)
    variable.position = rhs_v.value[0].position
    outcome = interpreter.value(variable, lhs_v.value)
    if outcome == Nothing():
        outcome = rhs_v.value[1]
    return outcome


def _addPublicVariableEventHandler(lhs_v, rhs_v, interpreter):
    interpreter.client.add_listening(lhs_v.value, rhs_v)


def _if_then_else_code(interpreter, condition, then, else_=None):
    """
    The equivalent Python code for a if-then-else SQF statement
    """
    assert(isinstance(condition, bool) and isinstance(then, Code))
    if condition:
        result = interpreter.execute_code(then)
    else:
        if else_ is not None:
            result = interpreter.execute_code(else_)
        else:
            result = Nothing()
    return result


def _if_then_else(if_instance, then_or_else, interpreter):
    condition = if_instance.condition.value
    if isinstance(then_or_else, Code):
        then = then_or_else
        else_ = None
    else:
        then = then_or_else.then
        else_ = then_or_else.else_

    return _if_then_else_code(interpreter, condition, then, else_)


def parse_switch(interpreter, code):
    conditions = []
    default_used = False

    for statement in code.base_tokens:
        base_tokens = statement.base_tokens

        # evaluate all the base_tokens, trying to obtain their values
        values = []
        for token in base_tokens:
            v = interpreter.value(token)
            values.append(v)

        if type(values[0]) != SwitchType:
            interpreter.exception(SQFParserError(
                statement.position, 'Switch code can only start with "case" or "default"'))

        if values[0].keyword == Keyword('default'):
            if default_used:
                interpreter.exception(SQFParserError(code.position, 'Switch code contains more than 1 `default`'))
            default_used = True
            assert(isinstance(values[0].result, Code))
            conditions.append(('default', values[0].result))
        else:
            case_condition = values[0].result
            if len(values) == 1:
                conditions.append((case_condition, None))
            else:
                assert (len(values) == 3 and values[1] == Keyword(':'))
                outcome_statement = values[2]
                conditions.append((case_condition, outcome_statement))

    return conditions


def execute_switch(interpreter, result, conditions):
    try:
        default = next(o for c, o in conditions if c == 'default')
    except StopIteration:
        default = None

    final_outcome = None

    execute_next = False
    for condition, outcome in conditions:
        if condition == 'default':
            continue
        condition_outcome = interpreter.value(condition)

        if outcome is not None and execute_next:
            final_outcome = interpreter.execute_code(outcome)
            break
        elif condition_outcome == result:
            if outcome is not None:
                final_outcome = interpreter.execute_code(outcome)
                break
            else:
                execute_next = True

    if final_outcome is None:
        if default is not None:
            final_outcome = interpreter.execute_code(default)
        else:
            final_outcome = Boolean(True)

    return final_outcome


def _foreach_loop(interpreter, code, elements):
    outcome = Nothing()
    for i, x in enumerate(elements):
        outcome = interpreter.execute_code(code, extra_scope={'_x': x, '_forEachIndex': Number(i)})
    return outcome


def _forvar_loop_code(interpreter, token_name, start, stop, step, code):
    outcome = Nothing()
    outcome.position = code.position

    for i in range(start, stop + 1, step):
        outcome = interpreter.execute_code(code, extra_scope={token_name: Number(i)})
    return outcome


def _forvar_loop(for_instance, code, interpreter):
    return _forvar_loop_code(interpreter,
                             for_instance.variable.value,
                             for_instance.from_.value,
                             for_instance.to.value, for_instance.step.value, code)


def _forspecs_loop_code(interpreter, start_code, stop_code, increment_code, do_code):
    outcome = Nothing()
    outcome.position = start_code.position

    interpreter.execute_code(start_code)
    while True:
        condition_outcome = interpreter.execute_code(stop_code)
        if condition_outcome.value is False:
            break

        outcome = interpreter.execute_code(do_code)
        interpreter.execute_code(increment_code)
    return outcome


def _forspecs_loop(forspec_type, do_code, interpreter):
    return _forspecs_loop_code(interpreter, forspec_type.array[0],
                               forspec_type.array[1], forspec_type.array[2], do_code)


def _while_loop(interpreter, condition_code, do_code):
    outcome = Nothing()
    while True:
        condition_outcome = interpreter.execute_code(condition_code)
        if condition_outcome.value is False:
            break
        outcome = interpreter.execute_code(do_code)
    return outcome


INTERPRETER_EXPRESSIONS = [
    TryCatchExpression(),

    ForEachExpression(lambda lhs, rhs, i: _foreach_loop(i, lhs, rhs.value)),
    WhileDoExpression(lambda lhs, rhs, i: _while_loop(i, lhs.condition, rhs)),

    ForFromToDoExpression(_forvar_loop),
    ForSpecDoExpression(_forspecs_loop),
    SwitchDoExpression(lambda lhs, rhs, i: execute_switch(i, lhs.result, parse_switch(i, rhs))),

    IfThenSpecExpression(lambda if_type, array, i: _if_then_else_code(i, if_type.condition.value, array.value[0], array.value[1])),
    IfThenElseExpression(_if_then_else),
    IfThenExpression(_if_then_else),
    IfThenExitWithExpression(),

    # params
    UnaryExpression(Keyword('params'), Array, Nothing, lambda rhs_v, i: i.add_params(rhs_v)),
    BinaryExpression(Type, Keyword('params'), Array, Nothing, lambda lhs_v, rhs_v, i: i.add_params(rhs_v)),

    # Unary
    UnaryExpression(Keyword('-'), Number, Number, Action(lambda x: -x.value)),
    UnaryExpression(Keyword('floor'), Number, Number, Action(lambda x: math.floor(x.value))),
    UnaryExpression(Keyword('reverse'), Array, Nothing, Action(lambda rhs_v: rhs_v.reverse())),
    # Binary
    BinaryExpression(Array, Keyword('set'), Array,
                     Nothing, Action(lambda lhs_v, rhs_v: lhs_v.set(rhs_v))),

    # Array related
    BinaryExpression(Array, Keyword('resize'), Number,
                     Nothing, Action(lambda lhs_v, rhs_v: lhs_v.resize(rhs_v.value))),
    UnaryExpression(Keyword('count'), Array, Number, Action(lambda x: len(x.value))),
    BinaryExpression(Type, Keyword('in'), Array, Boolean, Action(lambda x, array: x in array.value)),

    BinaryExpression(Array, Keyword('select'), Number, None, _select),
    BinaryExpression(Array, Keyword('select'), Boolean, None, _select),
    BinaryExpression(Array, Keyword('select'), Array, Array, _select_array),

    BinaryExpression(Array, Keyword('find'), Type, Number, Action(_find)),
    BinaryExpression(String, Keyword('find'), String, Number,
                     Action(lambda lhs_v, rhs_v: lhs_v.value.find(rhs_v.value))),

    BinaryExpression(Array, Keyword('pushBack'), Type, Number, Action(_pushBack)),
    BinaryExpression(Array, Keyword('pushBackUnique'), Type, Number, Action(_pushBackUnique)),
    BinaryExpression(Array, Keyword('append'), Array, Nothing, Action(lambda lhs_v, rhs_v: lhs_v.add(rhs_v.value))),

    UnaryExpression(Keyword('toArray'), String, Array,
                    Action(lambda rhs_v: [Number(ord(s)) for s in rhs_v.value])),
    UnaryExpression(Keyword('toString'), Array, String,
                    Action(lambda rhs_v: '"'+''.join(chr(s.value) for s in rhs_v.value)+'"')),

    # code and namespaces
    UnaryExpression(Keyword('call'), Code, None, lambda rhs_v, i: i.execute_code(rhs_v)),
    BinaryExpression(Type, Keyword('call'), Code, None, lambda lhs_v, rhs_v, i: i.execute_code(rhs_v, extra_scope={"_this": lhs_v})),

    BinaryExpression(Namespace, Keyword('setVariable'), Array, Nothing, _setVariable),

    BinaryExpression(Namespace, Keyword('getVariable'), String, None, _getVariableString),
    BinaryExpression(Namespace, Keyword('getVariable'), Array, None, _getVariableArray),

    BinaryExpression(String, Keyword('addPublicVariableEventHandler'), Code, None, _addPublicVariableEventHandler),

    BinaryExpression(Array, Keyword('+'), Array, Array, Action(lambda lhs_v, rhs_v: lhs_v.value + rhs_v.value)),
    BinaryExpression(Array, Keyword('-'), Array, Array, Action(_subtract_arrays)),

    BinaryExpression(String, Keyword('+'), String, String, Action(lambda lhs, rhs: lhs.container + lhs.value + rhs.value + lhs.container)),
]

for op in OP_COMPARISON:
    for lhs_rhs_type in [Number, String]:
        if lhs_rhs_type == Number or lhs_rhs_type == String and op in [Keyword('=='), Keyword('!=')]:
            INTERPRETER_EXPRESSIONS.append(ComparisonExpression(op, lhs_rhs_type))
for op in OP_ARITHMETIC:
    INTERPRETER_EXPRESSIONS.append(ArithmeticExpression(op))
for op in OP_LOGICAL:
    for rhs_type in (Boolean, Code):
        INTERPRETER_EXPRESSIONS.append(LogicalExpression(op, rhs_type))



from sqf.parser_types import ParserKeyword
from sqf.base_type import BaseType, ParserType, BaseTypeContainer


class Type(BaseType):
    """
    A type represents a type of variable. Every quantity that has a value is a type.
    """
    @property
    def is_undefined(self):
        """
        Represents whether its value is defined or not.
        """
        return self.value is None

    @property
    def value(self):
        return None


class ConstantValue(Type):
    """
    A constant (literal) value. For example, a number, a string, code.
    """
    def __init__(self, value=None):
        super().__init__()
        self._value = value

    @property
    def value(self):
        return self._value

    def __str__(self):
        return self.__class__.__name__


class Boolean(ConstantValue):
    def __init__(self, value=None):
        assert (value in (None, True, False))
        super().__init__(value)

    def __str__(self):
        if self.is_undefined:
            return 'undefined'
        if self._value:
            return 'true'
        else:
            return 'false'

    def __repr__(self):
        return 'B<%s>' % self


class String(ConstantValue):

    def __init__(self, value=None):
        self.container = None
        if value is not None:
            assert(isinstance(value, str))
            assert(value[0] == value[-1])
            assert (value[0] in ["'", '"'])
            self.container = value[0]
            super().__init__(value[1:-1])
        else:
            super().__init__(value)

    def __str__(self):
        if self.is_undefined:
            return "undefined"
        return "%s%s%s" % (self.container, self.value, self.container)

    def __repr__(self):
        return 's<%s>' % self


class Nothing(ConstantValue):
    """
    A type of unknown type
    """
    def __str__(self):
        return 'Nothing'

    def __repr__(self):
        return '<%s>' % self


class Anything(Type):
    """
    A type of unknown type
    """
    def __repr__(self):
        return '<Anything>'


class Number(ConstantValue):
    def __init__(self, value=None):
        assert(value is None or isinstance(value, (int, float)))
        super().__init__(value)

    def __str__(self):
        if self.is_undefined:
            return "undefined"
        if isinstance(self._value, int):
            return '%d' % self._value
        # todo: use a better representation of float
        return '%2.2f' % self._value

    def __repr__(self):
        return 'N%s' % self


class Variable(Type):
    """
    A variable that holds values. It has a name (e.g. "_x").
    """
    def __init__(self, name):
        super().__init__()
        self._name = name

    @property
    def name(self):
        return self._name

    @property
    def is_undefined(self):
        return False

    @property
    def is_global(self):
        return self.name[0] != '_'

    def __str__(self):
        return self._name

    def __repr__(self):
        return 'V<%s>' % self


class _Statement(BaseTypeContainer):
    def __init__(self, tokens, parenthesis=None, ending=None):
        assert (ending in (None, ',', ';'))
        assert (parenthesis in (None, '()', '[]', '{}'))
        assert (isinstance(tokens, list))
        for i, s in enumerate(tokens):
            assert(isinstance(s, (Type, Keyword, Preprocessor, Statement, ParserType)))

        self._parenthesis = parenthesis

        if self._parenthesis:
            tokens = [ParserKeyword(parenthesis[0])] + tokens + [ParserKeyword(parenthesis[1])]

        super().__init__(tokens)
        self._ending = None
        self.ending = ending

    def prepend(self, tokens):
        assert (isinstance(tokens, list))
        for i, s in enumerate(tokens):
            assert (isinstance(s, (Type, Keyword, Preprocessor, Statement, ParserType)))
        self._tokens = tokens + self._tokens

    @property
    def content(self):
        tokens = self.tokens.copy()
        if self.ending:
            del tokens[-1]
        if self.parenthesis:
            return tokens[1:-1]
        return tokens

    @staticmethod
    def is_base_token(token):
        # ignore tokens that are not relevant for the interpreter
        return not (isinstance(token, ParserType) or
                    isinstance(token, _Statement) and not token.parenthesis and not token.base_tokens)

    @property
    def ending(self):
        return self._ending

    @ending.setter
    def ending(self, ending):
        assert(ending in (None, ';', ','))
        if self._ending is not None:
            del self._tokens[-1]
        if ending is not None:
            self._tokens.append(ParserKeyword(ending))
        self._ending = ending

    def __len__(self):
        return len(self._tokens)

    def __getitem__(self, other):
        return self._tokens[other]

    def _as_str(self, func=str):
        return ''.join(func(item) for item in self._tokens)

    @property
    def parenthesis(self):
        return self._parenthesis


class Array(Type, BaseTypeContainer):

    def __init__(self, tokens=None):
        Type.__init__(self)
        if tokens is not None:
            self._values = tokens
        else:
            self._values = None
            tokens = []
        BaseTypeContainer.__init__(self, tokens)
        self.update_tokens()

    def update_tokens(self):
        self._tokens = [ParserKeyword('[')] + list(self._with_commas()) + [ParserKeyword(']')]

    def _with_commas(self):
        if self._values in [None, []]:
            return []
        it = iter(self._values)
        yield next(it)
        for x in it:
            yield ParserKeyword(',')
            yield x

    @property
    def is_undefined(self):
        return self._values is None

    def _as_str(self, func=str):
        if self.is_undefined:
            return '[undefined]'
        return ''.join(func(item) for item in self._tokens)

    def __len__(self):
        assert(not self.is_undefined)
        return len(self._values)

    def __getitem__(self, other):
        assert(not self.is_undefined)
        return self._values[other]

    @property
    def value(self):
        return self._values

    def __repr__(self):
        return '%s' % self._as_str(repr)

    def extend(self, index):
        assert (not self.is_undefined)
        new_tokens = [Nothing()] * (index - len(self._values) + 1)
        self._values += new_tokens
        self.update_tokens()

    def append(self, token):
        assert (not self.is_undefined)
        self._values.append(token)
        self.update_tokens()

    def resize(self, count):
        assert (not self.is_undefined)
        if count > len(self._values):
            self.extend(count - 1)
        else:
            self._values = self._values[:count]
            self.update_tokens()

    def reverse(self):
        assert (not self.is_undefined)
        self._values.reverse()
        self.update_tokens()

    def add(self, other):
        assert (not self.is_undefined)
        self._values += other
        self.update_tokens()

    def set(self, rhs_v):
        assert (not self.is_undefined)
        # https://community.bistudio.com/wiki/set
        assert(isinstance(rhs_v, Array))
        index = rhs_v.value[0].value
        value = rhs_v.value[1]

        if index >= len(self._values):
            self.extend(index)
        self._values[index] = value
        self.update_tokens()


class Statement(_Statement, BaseType):
    """
    The main class for holding statements. It is a BaseType because it can be nested, and
    it is a _Statement because it can hold elements.
    """
    def __init__(self, tokens, parenthesis=False, ending=None):
        if parenthesis:
            parenthesis = '()'
        else:
            parenthesis = None
        super().__init__(tokens, parenthesis, ending)

    def __repr__(self):
        return 'S<%s>' % self._as_str(repr)


class Code(_Statement, Type):
    """
    The class that holds (non-interpreted) code.
    """
    def __init__(self, tokens=None):
        Type.__init__(self)
        if tokens is not None:
            self._undefined = False
        else:
            self._undefined = True
            tokens = []
        _Statement.__init__(self, tokens, parenthesis='{}')

    @property
    def is_undefined(self):
        return self._undefined

    def __repr__(self):
        return '%s' % self._as_str(repr)


class Keyword(BaseType):
    def __init__(self, token):
        assert isinstance(token, str)
        super().__init__()
        self._token = token
        self._unique_token = self._token.lower()

    @property
    def value(self):
        return self._token

    @property
    def unique_token(self):
        return self._unique_token

    def __str__(self):
        return self._token

    def __repr__(self):
        return 'K<%s>' % self._token

    @property
    def _key(self):
        return self._unique_token,


class Namespace(Type):
    def __init__(self, token):
        assert isinstance(token, str)
        super().__init__()
        self._token = token
        self._unique_token = token.lower()

    @property
    def value(self):
        return self._token

    def __repr__(self):
        return 'NS<%s>' % self._token

    def __str__(self):
        return self._token

    @property
    def _key(self):
        return self._unique_token,


class Config(ConstantValue):
    pass


class Object(ConstantValue):
    pass


class File(Code):
    """
    Like code, but without parenthesis
    """
    def __init__(self, tokens):
        _Statement.__init__(self, tokens)

    def __repr__(self):
        return 'F<%s>' % self._as_str(repr)


class Preprocessor(Keyword):
    pass


class Script(ConstantValue):
    pass


class Control(ConstantValue):
    pass


class Group(ConstantValue):
    pass


class Display(ConstantValue):
    pass


class Side(ConstantValue):
    pass


class Task(ConstantValue):
    pass


class Location(ConstantValue):
    pass


class NetObject(ConstantValue):
    pass


class DiaryReport(ConstantValue):
    pass


class TeamMember(ConstantValue):
    pass
